package org.opengis.cite.iso19142.simple;

import java.net.URI;
import java.net.URL;
import java.util.logging.Level;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.validation.Validator;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.SuiteAttribute;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.ValidationUtils;
import org.opengis.cite.validation.SchematronValidator;
import org.opengis.cite.validation.XmlSchemaCompiler;
import org.testng.Assert;
import org.testng.ITestContext;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;
import org.xml.sax.SAXException;

/**
 * Contains tests regarding service metadata resources, especially the content
 * of the WFS capabilities document. The capabilities document must be
 * well-formed and schema-valid.
 */
@Test(groups = "simple")
public class ServiceMetadataTests extends BaseFixture {

    static final String ROOT_PACKAGE = "/org/opengis/cite/iso19142/";
    static final String SIMPLE_WFS_PHASE = "SimpleWFSPhase";
    private Schema wfsSchema;

    /**
     * Compiles the main WFS 2.0 schema. The resulting (thread-safe) Schema
     * object is stored in the test suite fixture as the value of the attribute
     * {@link SuiteAttribute#WFS_SCHEMA wfsSchema}.
     * 
     * @param testContext
     *            The test (group) context.
     */
    @BeforeClass(inheritGroups = true)
    public void compileWfsSchema(ITestContext testContext) {
        URL entityCatalog = this.getClass().getResource(
                ROOT_PACKAGE + "schema-catalog.xml");
        XmlSchemaCompiler xsdCompiler = new XmlSchemaCompiler(entityCatalog);
        Schema schema = null;
        try {
            schema = xsdCompiler.compileXmlSchema(URI.create(WFS2.SCHEMA_URI));
        } catch (SAXException ex) {
            TestSuiteLogger.log(Level.WARNING, "Failed to compile schema "
                    + WFS2.SCHEMA_URI, ex);
        }
        if (null != schema) {
            testContext.getSuite().setAttribute(
                    SuiteAttribute.WFS_SCHEMA.getName(), schema);
            this.wfsSchema = schema;
        } else {
            throw new RuntimeException("Failed to compile (internal) schema "
                    + WFS2.SCHEMA_URI);
        }
    }

    /**
     * Verifies that the WFS capabilities document is valid with respect to the
     * official {@code wfs.xsd} schema.
     */
    @Test
    public void capabilitiesDocIsXmlSchemaValid() {
        Validator validator = this.wfsSchema.newValidator();
        ETSAssert.assertSchemaValid(validator, new DOMSource(this.wfsMetadata,
                this.wfsMetadata.getDocumentURI()));
    }

    /**
     * Checks that the content of the WFS capabilities document reflects the
     * {@code Simple WFS} conformance class. The applicable rules are
     * incorporated into the {@value #SIMPLE_WFS_PHASE} phase of the Schematron
     * schema {@code wfs-capabilities-2.0.sch}.
     * 
     * @see "ISO 19142:2010, cl. A.1.1: Simple WFS"
     * @see "ISO 19142:2010, cl. A.2.23: Declaring conformance"
     */
    @Test
    public void capabilitiesDocCorrespondsToWfsSimple() {
        SchematronValidator validator = ValidationUtils
                .buildSchematronValidator("wfs-capabilities-2.0.sch",
                        SIMPLE_WFS_PHASE);
        DOMResult result = validator.validate(new DOMSource(this.wfsMetadata,
                this.wfsMetadata.getDocumentURI()));
        Assert.assertFalse(validator.ruleViolationsDetected(), ErrorMessage
                .format(ErrorMessageKeys.NOT_SCHEMA_VALID,
                        validator.getRuleViolationCount(),
                        XMLUtils.writeNodeToString(result.getNode())));
    }
}
