package org.opengis.cite.iso19142.transaction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.testng.annotations.AfterClass;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Tests the response to a Transaction request that includes one or more replace
 * actions.
 * 
 * @see "ISO 19142:2010, cl. 15.2.6: Replace action"
 */
@Test(groups = "transaction")
public class ReplaceTests extends TransactionFixture {

    /** List containing original representations of modified features */
    private List<Element> originalFeatures = new ArrayList<Element>();

    /**
     * Restores the WFS data store to its previous state by replacing all
     * modified features with their original representations.
     */
    @AfterClass
    public void restoreModifiedFeatures() {
        if (originalFeatures.isEmpty()) {
            return;
        }
        Document rspEntity = this.wfsClient.replace(originalFeatures,
                ProtocolBinding.ANY);
        String expr = String.format("//wfs:totalReplaced = '%d'",
                originalFeatures.size());
        Boolean result;
        try {
            result = (Boolean) XMLUtils.evaluateXPath(rspEntity, expr, null,
                    XPathConstants.BOOLEAN);
        } catch (XPathExpressionException xpe) {
            throw new RuntimeException(xpe);
        }
        if (!result) {
            String msg = String.format(
                    "%s: Failed to replace modified features.\n%s", getClass()
                            .getName(), XMLUtils.writeNodeToString(rspEntity));
            TestSuiteLogger.log(Level.WARNING, msg);
        }
    }

    /**
     * [{@code Test}] Submits a Transaction request to replace an existing
     * feature instance. The test is run for all supported protocol bindings
     * (except HTTP GET) and feature types. The response entity
     * (wfs:TransactionResponse) must be schema-valid and contain the
     * wfs:ReplaceResults element.
     * 
     * @param binding
     *            A supported transaction request binding.
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     * @see "ISO 19142:2010, cl. 15.3.3: TransactionSummary element"
     * @see "ISO 19142:2010, cl. 15.3.6: ReplaceResults element"
     */
    @Test(dataProvider = "binding+availFeatureType")
    public void replaceFeature(ProtocolBinding binding, QName featureType) {
        Document doc = wfsClient.getFeatureByType(featureType, 1, null);
        NodeList features = doc.getElementsByTagNameNS(
                featureType.getNamespaceURI(), featureType.getLocalPart());
        Element originalFeature = (Element) features.item(0);
        Element replacement = createReplacementFeature(originalFeature);
        List<Element> replacements = Arrays.asList(replacement);
        this.rspEntity = wfsClient.replace(replacements, binding);
        ETSAssert.assertXPath("//wfs:TransactionResponse/wfs:ReplaceResults",
                this.rspEntity, null);
        String xpath = String.format("//wfs:totalReplaced = '%d'",
                replacements.size());
        ETSAssert.assertXPath(xpath, this.rspEntity, null);
        originalFeatures.add(originalFeature);
    }

    /**
     * Creates a new feature representation by cloning and modifying the
     * original instance. The following standard GML properties are added (or
     * replaced if already present):
     * 
     * <ul>
     * <li>gml:description ("Lorem ipsum dolor sit amet.")</li>
     * <li>gml:identifier (random UUID value)</li>
     * </ul>
     * 
     * @param originalFeature
     *            The original feature instance (unmodified).
     * @return A new Element node representing the modified feature.
     */
    Element createReplacementFeature(Element originalFeature) {
        Element replacement = (Element) originalFeature.cloneNode(true);
        QName propName = new QName(Namespaces.GML, "identifier");
        Element identifier = XMLUtils.createElement(propName);
        identifier.setAttribute("codeSpace", "http://cite.opengeospatial.org/");
        identifier.setTextContent(UUID.randomUUID().toString());
        WFSRequest.insertGMLProperty(replacement, identifier);
        propName = new QName(Namespaces.GML, "description");
        Element desc = XMLUtils.createElement(propName);
        desc.setTextContent("Lorem ipsum dolor sit amet.");
        WFSRequest.insertGMLProperty(replacement, desc);
        return replacement;
    }
}
